# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                            calibration_internal.py
#               Runs the internal calibration step of the pipeline
# ******************************************************************************

"""
Obtains the external calibration zero-points by fitting models to a reference
catalog and comparing predicted S-PLUS magnitudes to instrumental magnitudes

The S-PLUS field is given as the first command line argument. Configurations
are set in the config file, given as the second command line argument.

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------

prepare_catalog_for_internal_calibration()
fit_sed_to_splus()
get_internal_calib_zero_points()
apply_external_zero_points()
plot_internal_zp()
combine_external_and_internal_zps()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
Ideally this script should only be run through the pipeline.py script.

Assumes that crossmatch.py has already been run for this field.

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 calibration_external.py *field_name* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

steps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(steps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field     = sys.argv[1]
conf_file = sys.argv[2]

conf = ut.pipeline_conf(conf_file)

################################################################################
# Get directories

field_path       = os.path.join(conf['run_path'], field)
crossmatch_path  = os.path.join(field_path, 'Crossmatch')

suffix = ut.calibration_suffix(conf)
calibration_path  = os.path.join(field_path, f'Calibration_{suffix}')

external_cal_path = os.path.join(calibration_path, 'external')
external_plot_path = os.path.join(external_cal_path, 'plots')

stloc_cal_path = os.path.join(calibration_path, 'stlocus')
stloc_plot_path = os.path.join(stloc_cal_path, 'plots')

internal_cal_path = os.path.join(calibration_path, 'internal')
internal_plot_path = os.path.join(internal_cal_path, 'plots')

log_path = os.path.join(calibration_path, 'logs')


################################################################################
# Initiate log file

ut.makedir(calibration_path)
ut.makedir(internal_cal_path)
ut.makedir(internal_plot_path)
ut.makedir(log_path)

log_file_name = os.path.join(log_path, 'calibration_internal.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(calibration_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")

################################################################################
# Begin script

# ***************************************************
#    Copy instrumental magnitudes catalog
# ***************************************************


def prepare_catalog_for_internal_calibration():

    """
    Copy S-PLUS instrumental magnitudes to external calibration path
    """

    print("")
    ut.printlog(('********** '
                 'Preparing pre-internal calibration S-PLUS magnitudes catalog '
                 '**********'),
                 log_file)
    print("")

    if 'calibration_stloc' in conf['run_steps']:
        catalog_name = f"{field}_mag_stlocus.cat"
        catalog_file = os.path.join(stloc_cal_path, catalog_name)

    else:
        catalog_name = f"{field}_mag_ext.cat"
        catalog_file = os.path.join(external_cal_path, catalog_name)

    save_name = f"{field}_pre_internal.cat"
    save_file = os.path.join(internal_cal_path, save_name)

    if not os.path.exists(save_file):

        # Copy the last calibrated catalog
        cmd = f"cp {catalog_file} {save_file}"
        ut.printlog(cmd, log_file)
        os.system(cmd)

    else:
        ut.printlog(f"File {save_name} already created", log_file)


prepare_catalog_for_internal_calibration()

# ***************************************************
#    Fit SEDs to reference catalog
# ***************************************************


def fit_sed_to_splus():

    """
    Fits SEDs to the reference catalog magnitudes to predict the S-PLUS
    calibrated magnitudes
    """

    print("")
    ut.printlog(('********** '
                 'Fitting SEDs to reference '
                 '**********'),
                 log_file)
    print("")

    models_file = conf['path_to_models']

    splus_mag_catalog = f"{field}_pre_internal.cat"
    splus_mag_file    = os.path.join(internal_cal_path, splus_mag_catalog)

    bayesian_fitting = conf['model_fitting_bayesian']
    ebv_mode = conf['model_fitting_ebv_cut']

    ref_mag_cols  = conf['internal_sed_fit']

    #pred_mag_cols = None
    pred_mag_cols = conf['internal_sed_pred'] + conf['external_sed_fit']

    save_name = f"{field}_splus_sed_fit.cat"
    save_file = os.path.join(internal_cal_path, save_name)

    if not os.path.exists(save_file):
        ut.sed_fitting(models        = models_file,
                       catalog       = splus_mag_file,
                       save_file     = save_file,
                       ref_mag_cols  = ref_mag_cols,
                       pred_mag_cols = pred_mag_cols,
                       bayesian      = bayesian_fitting,
                       ebv_mode      = ebv_mode)

        ut.printlog(f"Generated SED fit catalog {save_file}", log_file)
    else:
        ut.printlog(f"SED fit catalog {save_name} already exists", log_file)


fit_sed_to_splus()


# ***************************************************
#    Get external calibration zero points
# ***************************************************

def get_internal_calib_zero_points():

    """
    Obtains the S-PLUS internal calibration zero-points
    """

    print("")
    ut.printlog(('********** '
                 'Getting internal calibration ZPs '
                 '**********'),
                 log_file)
    print("")

    catalog_name = f"{field}_splus_sed_fit.cat"
    catalog_file = os.path.join(internal_cal_path, catalog_name)

    save_name = f"{field}_internal_tmp.zp"
    save_file = os.path.join(internal_cal_path, save_name)

    if not os.path.exists(save_file):
        ut.zp_estimate(catalog = catalog_file,
                       save_file = save_file,
                       filters_list = conf['internal_sed_pred'],
                       mag_cut   = conf['zp_fitting_mag_cut'])

        ut.printlog(f"Created zp file {save_file}", log_file)

    else:
        ut.printlog(f"ZP file {save_name} already exists", log_file)


get_internal_calib_zero_points()


# ***************************************************
#    Add offset to S-PLUS reference calibration
# ***************************************************

def add_offset_to_splus_reference_calibration():
    """
    Adds offset to S-PLUS reference calibration
    """

    if conf['offset_include_after'].lower() == "internal":
        print("")
        ut.printlog(('********** '
                     'Adding offset to S-PLUS reference calibration '
                     '**********'),
                    log_file)
        print("")

    internal_zp_name = f"{field}_internal_tmp.zp"
    internal_zp_file = os.path.join(internal_cal_path, internal_zp_name)

    save_name = f"{field}_internal.zp"
    save_file = os.path.join(internal_cal_path, save_name)

    if not os.path.exists(save_file):

        if conf['offset_include_after'].lower() == "internal":
            offset_zp_name = f'offset_refcalib.zp'
            offset_zp_file = os.path.join(internal_cal_path, offset_zp_name)

            cmd = f"cp {conf['offset_to_splus_refcalib']} {offset_zp_file}"
            ut.printlog(cmd, log_file)
            os.system(cmd)

            zp_file_list = [internal_zp_file, offset_zp_file]

            filters = []
            for filt in conf['filters']:
                filters.append(f"SPLUS_{filt}")

            # Combine zero-points
            ut.zp_add(zp_file_list=zp_file_list,
                      save_file=save_file,
                      filters=filters,
                      inst_zp=0)

        else:
            cmd = f"cp {internal_zp_file} {save_file}"
            os.system(cmd)

        ut.printlog(f"Created file {save_file}", log_file)

    else:
        ut.printlog(f"File {save_name} already exists", log_file)


add_offset_to_splus_reference_calibration()


# ***************************************************
#    Apply Zero Points
# ***************************************************


def apply_internal_zero_points():

    """
    Applies internal zero-points to mag_inst catalog
    """

    print("")
    ut.printlog(('********** '
                 'Applying internal zero-points to mag_inst catalog '
                 '**********'),
                 log_file)
    print("")

    catalog_name = f"{field}_splus_sed_fit.cat"
    catalog_file = os.path.join(internal_cal_path, catalog_name)

    zp_name = f"{field}_internal.zp"
    zp_file = os.path.join(internal_cal_path, zp_name)

    save_name = f"{field}_mag_int.cat"
    save_file = os.path.join(internal_cal_path, save_name)

    if not os.path.exists(save_file):
        ut.zp_apply(catalog   = catalog_file,
                    save_file = save_file,
                    zp_file   = zp_file)

        ut.printlog(f"Created file {save_file}", log_file)

    else:
        ut.printlog(f"Catalog {save_name} already exists", log_file)


apply_internal_zero_points()

# ***************************************************
#          Plot internal ZPs fitting
# ***************************************************


def plot_internal_zp():

    """
    Makes diagnostic plots of the external zp fitting process
    """

    print("")
    ut.printlog(('********** '
                 'Plotting internal zp-fitting process '
                 '**********'),
                 log_file)
    print("")

    sed_fit_name = f"{field}_splus_sed_fit.cat"
    sed_fit_file = os.path.join(internal_cal_path, sed_fit_name)

    zp_name = f"{field}_internal.zp"
    zp_file = os.path.join(internal_cal_path, zp_name)

    for filt in conf['internal_sed_pred']:
        save_name = f"{field}_{filt}_int.png"
        save_file = os.path.join(internal_plot_path, save_name)

        if not os.path.exists(save_file):
            ut.plot_zp_fitting(sed_fit_file = sed_fit_file,
                               zp_file      = zp_file,
                               save_file    = save_file,
                               filt         = filt,
                               mag_cut      = conf['zp_fitting_mag_cut'])

            ut.printlog(f"Saved plot {save_file}", log_file)

        else:
            ut.printlog(f"Plot {save_name} already made", log_file)


plot_internal_zp()


# ***************************************************
#    Combine external + stlocus + internal zps
# ***************************************************

def combine_external_and_internal_zps():
    """
    Combines the external and internal .zp files into a single .zp file
    """

    print("")
    ut.printlog(('********** '
                 'Combining external and internal .zp files '
                 '**********'),
                log_file)
    print("")

    if 'calibration_stloc' in conf['run_steps']:
        ext_zp_name = f"{field}_external_and_stlocus.zp"
        ext_zp_file = os.path.join(stloc_cal_path, ext_zp_name)
    else:
        ext_zp_name = f"{field}_external.zp"
        ext_zp_file = os.path.join(external_cal_path, ext_zp_name)

    int_zp_name = f"{field}_internal.zp"
    int_zp_file = os.path.join(internal_cal_path, int_zp_name)

    save_name = f"{field}_external_and_internal.zp"
    save_file = os.path.join(internal_cal_path, save_name)

    filters = []
    for filt in conf['filters']:
        filters.append(f"SPLUS_{filt}")

    if not os.path.exists(save_file) or True:

        ut.zp_add(zp_file_list=[ext_zp_file, int_zp_file],
                  save_file=save_file,
                  filters=filters)

        ut.printlog(f"Created file {save_file}", log_file)

    else:
        ut.printlog(f"Catalog {save_name} already exists", log_file)


combine_external_and_internal_zps()



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                               correction_xy.py
#                  ad-hoc zero-point inhomogeneity correction
# ******************************************************************************


"""
Applies ad-hoc zero-point inhomogeneity corrections to the aperture photometry
catalogues.

The correction maps are saved as numpy .npy arrays, with bins defined in the
configuration file

The S-PLUS field is given as the first command line argument.

The set of filters, location of correction maps and definition of the bins,
must be set in the configuration file given as the second command line argument

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------

create_xycorr_paths()
apply_xycorr_single_catalogs()
apply_xycorr_dual_catalogs()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
Ideally this script should only be run through the pipeline.py script.

Assumes that photometry_single.py and/or photometry_dual.py has already been
run for this field.

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 correction_xy.py *field_name* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

steps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(steps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field     = sys.argv[1]
conf_file = sys.argv[2]

conf = ut.pipeline_conf(conf_file)

################################################################################
# Get directories

field_path           = os.path.join(conf['run_path'], field)
photometry_path      = os.path.join(field_path, 'Photometry')

dual_path            = os.path.join(photometry_path, 'dual')
dual_catalogs_path   = os.path.join(dual_path, 'catalogs')
dual_xycorr_path     = os.path.join(dual_path, 'xy_correction')

single_path          = os.path.join(photometry_path, 'single')
single_catalogs_path = os.path.join(single_path, 'catalogs')
single_xycorr_path   = os.path.join(single_path, 'xy_correction')

psf_path          = os.path.join(photometry_path, 'psf')
psf_catalogs_path = os.path.join(psf_path, 'catalogs')
psf_xycorr_path   = os.path.join(psf_path, 'xy_correction')

log_path             = os.path.join(photometry_path, 'logs')

################################################################################
# Initiate log file

ut.makedir(log_path)

log_file_name = os.path.join(log_path, 'xy_correction.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(photometry_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")

################################################################################
# Log configuration

ut.printlog("PSF photometry parameters:", log_file)

single_mode_params = ['sex_XY_correction', 'XY_correction_maps_path',
                      'XY_correction_xbins', 'XY_correction_ybins']

for param in single_mode_params:
    try:
        ut.printlog(f"{param}: {conf[param]}", log_file)
    except KeyError:
        ut.printlog(f"{param}: NONE", log_file)


################################################################################
# Begin script

# ***************************************************
#    Create paths
# ***************************************************

def create_xycorr_paths():

    print("")
    ut.printlog(('********** '
                 'Generating XY corrections paths '
                 '**********'),
                 log_file)
    print("")

    # Create single mode photometry path
    if 'photometry_dual' in conf['run_steps']:
        ut.makedir(dual_xycorr_path)

    if 'photometry_single' in conf['run_steps']:
        ut.makedir(single_xycorr_path)

    if 'photometry_psf' in conf['run_steps']:
        ut.makedir(psf_xycorr_path)

create_xycorr_paths()


# ***************************************************
#    Apply XY correction to single mode catalogs
# ***************************************************


def apply_xycorr_single_catalogs():

    print("")
    ut.printlog(('********** '
                 'Applying XY corrections to single mode catalogs '
                 '**********'),
                 log_file)
    print("")

    for filt in conf['filters']:

        catalog_name = f"sex_{field}_{filt}_single.fits"
        catalog      = os.path.join(single_catalogs_path, catalog_name)

        xycorr_catalog_name = f"sex_{field}_{filt}_single_xycorr.fits"
        save_file           = os.path.join(single_xycorr_path,
                                           xycorr_catalog_name)

        xycorr_map_name = f"SPLUS_{filt}_offsets_grid.npy"
        xycorr_map_file = os.path.join(conf['XY_correction_maps_path'],
                                       xycorr_map_name)

        if not os.path.exists(save_file):

            ut.apply_xy_correction(catalog   = catalog,
                                   save_file = save_file,
                                   map_file  = xycorr_map_file,
                                   xbins     = conf['XY_correction_xbins'],
                                   ybins     = conf['XY_correction_ybins'])

            ut.printlog(("Generated single mode XY corrected catalog for field "
                         f"{field}, filter {filt}"), log_file)

        else:
            ut.printlog(("Single mode XY corrected catalog already created for "
                         f"field {field}, filter {filt}"), log_file)


if 'photometry_single' in conf['run_steps']:
    apply_xycorr_single_catalogs()


# ***************************************************
#    Apply XY correction to dual mode catalogs
# ***************************************************


def apply_xycorr_dual_catalogs():
    print("")
    ut.printlog(('********** '
                 'Applying XY corrections to dual mode catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:

        catalog_name = f"sex_{field}_{filt}_dual.fits"
        catalog = os.path.join(dual_catalogs_path, catalog_name)

        xycorr_catalog_name = f"sex_{field}_{filt}_dual_xycorr.fits"
        save_file = os.path.join(dual_xycorr_path,
                                 xycorr_catalog_name)

        xycorr_map_name = f"SPLUS_{filt}_offsets_grid.npy"
        xycorr_map_file = os.path.join(conf['XY_correction_maps_path'],
                                       xycorr_map_name)

        if not os.path.exists(save_file):

            ut.apply_xy_correction(catalog=catalog,
                                   save_file=save_file,
                                   map_file=xycorr_map_file,
                                   xbins=conf['XY_correction_xbins'],
                                   ybins=conf['XY_correction_ybins'])

            ut.printlog(("Generated dual mode XY corrected catalog for field "
                         f"{field}, filter {filt}"), log_file)

        else:
            ut.printlog(("Dual mode XY corrected catalog already created for "
                         f"field {field}, filter {filt}"), log_file)


if 'photometry_dual' in conf['run_steps']:
    apply_xycorr_dual_catalogs()


# ***************************************************
#    Apply XY correction to psf mode catalogs
# ***************************************************


def apply_xycorr_psf_catalogs():
    print("")
    ut.printlog(('********** '
                 'Applying XY corrections to psf catalogs '
                 '**********'),
                log_file)
    print("")

    for filt in conf['filters']:

        catalog_name = f"{field}_{filt}_psf.cat"
        catalog = os.path.join(psf_catalogs_path, catalog_name)

        xycorr_catalog_name = f"{field}_{filt}_psf_xycorr.cat"
        save_file = os.path.join(psf_xycorr_path,
                                 xycorr_catalog_name)

        xycorr_map_name = f"SPLUS_{filt}_offsets_grid.npy"
        xycorr_map_file = os.path.join(conf['XY_correction_maps_path'],
                                       xycorr_map_name)

        if not os.path.exists(save_file):

            ut.apply_xy_correction_psf(catalog=catalog,
                                       save_file=save_file,
                                       map_file=xycorr_map_file,
                                       xbins=conf['XY_correction_xbins'],
                                       ybins=conf['XY_correction_ybins'])

            ut.printlog(("Generated psf XY corrected catalog for field "
                         f"{field}, filter {filt}"), log_file)

        else:
            ut.printlog(("PSF XY corrected catalog already created for "
                         f"field {field}, filter {filt}"), log_file)


if 'photometry_psf' in conf['run_steps']:
    apply_xycorr_psf_catalogs()


#====================================================================================================================#

# For general purposes
import numpy as np
import os, sys

# To handle .fits files and create tables
from astropy.io import fits
from astropy.table import Table, vstack

# To interpolate the filters
from scipy.interpolate import interp1d

# To add the extinction in the flux
from dust_laws import CCM

# To calculate the prior probability of the models
from prior import PriorCalculator

# Just for type hinting
from typing import Callable

#====================================================================================================================#


class SpectralModel:

    """
    This class is used as a base for generating models (from spectral libraries), and it is meant to be used as a
    parent class*, given that each spectral library has its own way of loading the wavelength and flux values from
    the files (and also obtaining the parameters from the name of the file).
    The E(B-V) values that you want the model to use need to be passed in the constructor (otherwise it will just use 0).

    * THIS CLASS WONT WORK WITHOUT A LOAD FUNCTION TO GET THE FLUXES AND WAVELENGHTS OUT OF THE MODEL FILES

    Methods:
    ------------
    self.fill(path_to_filters: str, *args: list[str]):
        Fills the data table with the stellar parameters and magnitudes of each model on the given filters.
    
    self.add_prior(reference: tuple, bins: dict, regbins: bool, **kwargs):
        Adds a prior column, to the the data table, calculated based on the given reference catalog.
    
    self.interpolate(params: list[tuple], restrinct: function):
        Iterpolates the original fluxes to get a thinner grid of models.

    self.save(path: str, fmt: str):
        Saves the data table into a file of the given format.
    
    self.remove(col: str, value):
        Removes every line in the catalog that has value in the column col.

    self.copy(model: SpectralModel):
        Loads the data that another SpectralModel object has already loaded and interpolated.

    self.resampler(x_old, y_old, x_new):
        Used for interpolating filters.
    
    self.synmag(wl, flux, filter_curve):
        Convolutes the transmission curve of the filter with the SED to get the AB magnitude (on that filter).
    """

    def __init__(self, ebvs = [0], order = ["Teff", "logg", "FeH", "aFe"]):
        self.ebvs           = np.array(ebvs)
        self.save_prior     = False
        self.order          = order
        self.cols           = ["model_id", "EB_V", "Teff", "logg", "FeH", "aFe"]


    def fill(self, path_to_filters, *args):

        """
        Fills the self.synthetic_data table with the magnitudes of every model on the filters given.

        Parameters:
        ------------
        path_to_filters: str
            Path to the folder containing the filter transmitance curves.
        *args: list[str]
            Any quantity of arrays with the names of the filters.
        """
        
        # Initializes the filter dictionary
        self.filters = {}

        # Gets the filter curves
        for arg in args:
            for filt in arg:
                self.filters[filt] = np.genfromtxt(path_to_filters + filt + '.dat').transpose()
                self.cols.append(filt)
        
        print()
        # Creates an empty array for the data table
        self.synthetic_data = np.empty((len(self.templates) * len(self.ebvs), 6 + len(self.filters)))

        for e, extinction in enumerate(self.ebvs):
            for i, template in enumerate(self.templates):

                # Gets the name of the model
                file = self.files[i]

                # Message to the user
                message = f"Processing \u001b[33;1m{file}\u001b[0m with E(B-V) = \u001b[38;5;204m{extinction}\u001b[0m"
                sys.stdout.write(f'\u001b[32;1m>>>\u001b[0m {message} \r')
                sys.stdout.flush()

                # Gets the wavelenghts and fluxes from the model dictionary
                wl, flux = self.models_dict[file]['lin_wl'], self.models_dict[file]['flux']

                # Calculates the flux with extinction
                av = 3.1 * extinction
                a_lambda = av * CCM(wl)
                flux_with_dust = flux * 10 ** (-0.4 * a_lambda)

                # Gets the stellar parameters of the model
                Teff, logg, FeH, aFe = template

                # Writes data of this model into the table
                index_data                          = i + e * len(self.templates)
                self.synthetic_data[index_data, 0]  = index_data
                self.synthetic_data[index_data, 1]  = extinction
                self.synthetic_data[index_data, 2]  = Teff
                self.synthetic_data[index_data, 3]  = logg
                self.synthetic_data[index_data, 4]  = FeH
                self.synthetic_data[index_data, 5]  = aFe
                self.synthetic_data[index_data, 6:] = np.array([self.synmag(wl           = wl,
                                                                            flux         = flux_with_dust,
                                                                            filter_curve = self.filters[filt])
                                                                            for filt in self.filters])


    def add_prior(self, reference, bins, regbins = False, **kwargs):
        
        """
        Adds a column with the values of the prior probabilities of each model
        (Used AFTER the self.fill(...) function).

        Parameters:
        ------------
        reference: tuple
            Ordered* tuple with the path to the reference, the column names of
            the parameters on the reference and the format of the file.
        bins: dict
            Ordered* dictionary with the parameter names and a list with the bins
            of that parameter to be used.
            It is important to add the last parameter with an empty list (even it
            not having the need to have bins, we need the name of the column to 
            separate and order it before calling the calculator with the templates).
        regbins: bool (optinal)
            If the bins are regular (evenly spaced), in which case the "bins"
            parameter should contain tuples of the type (N, r0, r1) instead of the
            actual bins.
        **kwargs:
            bw: str ('scott' / 'silverman') or float
                bandwidth method/factor that's going to be used on the KDEs.
            normalize_by: str
                Normalization method (for now, 'max' or 'sum').
        
        * by "ordered" we mean from the most independent parameter to the most 
          nested (if the calculation is meant to be P(q0)P(q1|q0)...P(qn|q(n-1)))
          the order should be q0 -> qn.
        """

        self.cols.append("prior")
        print()

        # (For the self.save)
        self.save_prior = True

        # Decompose the bins into two parts:
        pos         = []
        real_bins   = []
        for p in bins:
        
            # The positions (in order) of the template columns
            pos.append(self.order.index(p))
        
            # The actual bins (or tuples to the regular_bins constructor)
            real_bins.append(bins[p])

        # If the user chooses to use regular bins
        if regbins:
            
            # We initialize the PriorCalculator without passing in the bins
            self.p = PriorCalculator(**kwargs)

            # And construct them with the regular_bins method
            self.p.regular_bins(real_bins[:-1])
        
        # If not
        else:

            # Simply pass the bins to the constructor
            self.p = PriorCalculator(real_bins[:-1], **kwargs)
        
        # Reads the reference and creates the KDEs with it
        self.p.interpret(*reference, lastBound = real_bins[-1][-2:])

        # Creates a new list of tuples but with only the parameters needed and in the right order
        tmps = list(zip(*self.templates))
        params_in_order = [tmps[i] for i in pos]
        params_in_order = list(zip(*params_in_order))

        # Generates a prior column to the templates passed in (is the same for every ebv)
        prior_column = self.p.prior_column(params_in_order) * len(self.ebvs)

        # Adds the prior column to the data table
        self.synthetic_data = np.insert(self.synthetic_data, len(self.synthetic_data[0]), [prior_column], axis=1)


    def interpolate(self, params, restrict = None):

        '''
        Interpolates the original model templates* to generate a thinner grid.

        Parameters:
        ------------
        params: list[tupple]
            Array of quadruples (Teff, logg, [Fe/H], [a/Fe]) to (try*) adding to the grid.
        restrict: function (optional)
            Function that receives a quadruplet of parameters and return a boolean (should use or not).

        * For now, this implementation just works for coelho14 (because of the wavelenght array).
        ** Some of the quadrupples are going to be ignored if the interpolation for them doesn't work
          (if they are out of the convex hull) .
        '''
        
        # Creates empty arrays for the important quantities
        table   = [] # Stellar parameters
        waves   = [] # Wavelenghts
        fluxes  = [] # Fluxes

        # For every file of the original models
        for i, file in enumerate(self.files):
            
            # Takes the (already saved) wavelenght and flux arrays
            wl, flux = self.models_dict[file]['lin_wl'], self.models_dict[file]['flux']
            
            # Saves the stellar parameters in a Table instance
            a = Table(np.array(self.templates[i]), names=["T", "g", "Z", "alpha"])
            
            # Append everything to the arrays
            waves.append(wl)
            fluxes.append(flux)
            table.append(a)

        # With those quantities, uses paintbox to create an interpolator
        star = pb.ParametricModel(waves[0], vstack(table), np.array(fluxes))

        # Filters the parameters with the restriction passed in
        filtered_pars = list(filter(restrict, params)) if (restrict != None) else params

        print()

        # For every requested stellar parameter quadruple
        for theta in filtered_pars:

            # Uses the interpolator to calculate the flux array
            flux = star(theta)

            # If the quadruple is not already on the templates (and the interpolation works)
            if (not theta in self.templates) and (0 not in flux):
            #if (not theta in self.templates):

                # Gives a name to the new model 
                file = "interpolated_T{}_g{}_z{}_a{}".format(*theta)

                # Message to the user
                message = f"Adding \u001b[33;1m{file}\u001b[0m to the templates"
                sys.stdout.write(f'\u001b[32;1m>>>\u001b[0m {message}\r')
                sys.stdout.flush()

                # Appends, both the quadruple to the templates and the file name to self.files
                self.templates.append(theta)
                self.files.append(file)

                # Adds the model information to the models_dict (calculating the flux with the interpolator)
                self.models_dict[file]           = {}
                self.models_dict[file]['lin_wl'] = np.array(waves[0])
                self.models_dict[file]['flux']   = flux


    def save(self, path, fmt, form = [0, 3, 5]):

        """
        Saves the data table into a file of the given format.

        Parameters:
        ------------
        path: str
            Path to save model catalog
        fmt: str
            Format that the catalog should be saved as (e.g., 'ascii').
        form: list[int] (optional)
            Number of digits of the index, the stellar parameters (+ magnitudes)
            and priors, in this order.
        """

        # Message to the user
        message = f"Saving model catalog at \u001b[33;1m{path}\u001b[0m (\u001b[38;5;204m{fmt}\u001b[0m format)"
        print(f'\n\u001b[32;1m>>>\u001b[0m {message}\n')


        # Saves as a .cat file
        if (fmt == "ascii"):

            # Creates the file
            with open(path, 'w') as f:
                
                # Format of the index / magnitudes
                ffmmt = [f'%.{form[0]}f'] + [f'%.{form[1]}f' for _ in range(len(self.synthetic_data[0]) - 1)]

                # Writes the stellar parameters names into the header
                f.write('# model_id EB_V Teff logg FeH aFe')
                
                # Writes the filter names into the header
                for filter in self.filters:
                    f.write(' {}'.format(filter))
                
                # If self.add_prior(...) was called, adds 'prior' to the header
                if self.save_prior:
                    f.write(' prior')
                    ffmmt[-1] = f'%.{form[2]}f'
                
                # Puts the data into the file
                f.write('\n')
                np.savetxt(f, self.synthetic_data, fmt=ffmmt, delimiter = ' ')


    def copy(self, model):

        """
        Loads the data that another SpectralModel object has already loaded and interpolated.

        Parameters:
        ------------
        model: SpectralModel instance
            Model from which whit function will copy the loaded data from.
        """

        self.templates      = [theta[:] for theta in model.templates]
        self.models_dict    = model.models_dict.copy()
        self.files          = model.files[:]


    def remove(self, col, value, error = 0):

        C = self.cols.index(col)

        if error != 0:
            del_indexes = [i for i in range(self.synthetic_data.shape[0])
                            if abs(self.synthetic_data[i, C] - value) < error]    
        else:
            del_indexes = [i for i in range(self.synthetic_data.shape[0])
                            if self.synthetic_data[i, C] == value]

        self.synthetic_data = np.delete(self.synthetic_data, del_indexes, 0)


    def resampler(self, x_old, y_old, x_new):

        " This function is called by synmag to interpolate filters"
        
        interp = interp1d(x_old, y_old, bounds_error=False, fill_value=(0., 0.))
        y_new  = interp(x_new)
        
        return y_new


    def synmag(self, wl, flux, filter_curve):

        """
        Convolutes the filter trasmitance curve with the SED to give the AB magnitude on that filter.

        Parameters
        ------------
        wl: array-like
            Wavelengths.
        flux: array-like
            Fluxes.
        filter_curve: str or array-like
            Filter curve, can either be a 2xn array where position 0 a vector of wavelengths and position 1 is a vector
            of trasmitances or a string containing the path to a filter file.

        Returns
        ------------
        m_ab: array-like
            Synthetic AB magnitude.
        """

        # Reading filter if needed:
        if type(filter_curve) is str:
            wl_filter, transmitance = np.genfromtxt(filter_curve).transpose()
        else:
            wl_filter, transmitance = filter_curve[0], filter_curve[1]

        # Resampling filter and spectrum to 1 Angstrom intervals:
        wl_new = np.arange(np.round(wl_filter[0]) - 5, np.round(wl_filter[-1]) + 5)

        transmitance = self.resampler(wl_filter, transmitance, wl_new)
        flux = self.resampler(wl, flux, wl_new)

        # Calculating the magnitude:
        m_ab = -2.5 * np.log10(np.trapz(flux * transmitance * wl_new, dx=1) / np.trapz(transmitance / wl_new, dx=1)) - 2.41

        return m_ab


#====================================================================================================================#


class Coelho_2014(SpectralModel):

    """
    This class is a child class of Model, it uses all the original methods from it, but it has a Load function to
    load the data from the model files.

    The models to be used with this class are described on:
        Coelho P. R. T., 2014, MNRAS, 440, 1027

    Methods:
    ------------
    self.load(path: str):
        Loads the wavelenghts, fluxes and stellar parameters from the models folder passed as argument.
    """

    def load(self, path):

        # Looks into the directory passed for the templates
        file_list = os.listdir(path)

        # Reads a reference model to get wavelengths
        model   = fits.open(path + file_list[0])
        log_wl  = np.array([model[0].header['CRVAL1'] + i * model[0].header['CDELT1']
                            for i in range(model[0].header['NAXIS1'])])
        lin_wl  = 10**log_wl

        # Initializes all the containers
        self.models_dict = { }
        self.templates   = [ ]
        self.files       = [ ]

        print()
        for i, file in enumerate(file_list):

            # Reads the model
            model = fits.open(path + file)

            # Removes the extension from the name of the file
            file  = file.replace('.fits', '')

            # Message to the user
            message = f'Reading model \u001b[38;5;204m{i + 1}\u001b[0m of \u001b[38;5;204m{len(file_list)}\u001b[0m'
            sys.stdout.write(f'\u001b[32;1m>>>\u001b[0m {message} : \u001b[33;1m{file}\u001b[0m\r')
            sys.stdout.flush()

            # Fills the model dictionary
            self.files.append(file)
            self.models_dict[file]            = {}
            self.models_dict[file]['lin_wl']  = lin_wl
            self.models_dict[file]['flux']    = model[0].data

            # Gets the parameters from the file name
            params = file.split("_")
            
            Teff = float(params[0][1:])
            logg = float(params[1][1:])
            
            FeH_sign = params[2][0]
            FeH_sign = +1 if FeH_sign == 'p' else -1
            FeH = FeH_sign * float(params[2][1] + '.' + params[2][2])
            
            aFe_sign = params[2][3]
            aFe_sign = +1 if aFe_sign == 'p' else -1
            aFe = aFe_sign * float(params[2][4] + '.' + params[2][5])

            # Appends, to the template array, the stellar parameters of the model
            self.templates.append([Teff, logg, FeH, aFe])


#====================================================================================================================#


class Castelli_Kurucz_2003(SpectralModel):

    """
    This class is a child class of Model, it uses all the original methods from it, but it has a Load function to
    load the data from the model files.

    The models to be used with this class are described on:
        Castelli F., Kurucz R. L., 2003, in Piskunov N., Weiss W. W., Gray
            D. F., eds, Vol. 210, Modelling of Stellar Atmospheres. p. A20
            (arXiv:astro-ph/0405087)

    Methods:
    ------------
    self.load(path: str):
        Loads the wavelenghts, fluxes and stellar parameters from the models folder passed as argument.
    """

    def load(self, path):

        # logg values to be used in the models
        kurucz_g = ['g00', 'g05', 'g10', 'g15',
                    'g20', 'g25', 'g30', 'g35',
                    'g40', 'g45', 'g50']

        # Looks into the directory passed for the templates
        file_list = os.listdir(path)

        # Initializes all the containers
        self.models_dict = { }
        self.templates   = [ ]
        self.files       = [ ]

        print()        
        for i, file in enumerate(file_list):

            # Reads the model
            model = fits.open(path + file)

            # Removes the extension from the name of the file
            file  = file.replace('.fits', '')

            for g in kurucz_g:

                # Message to the user
                a = i * len(kurucz_g) + 1
                b = len(file_list) * len(kurucz_g)
                message = f'Reading model \u001b[38;5;204m{a}\u001b[0m of \u001b[38;5;204m{b}\u001b[0m'
                sys.stdout.write(f'\u001b[32;1m>>>\u001b[0m {message} : \u001b[33;1m{file}_{g}\u001b[0m\r')
                sys.stdout.flush()

                if len(set(model[1].data[g])) > 1: # models without data are set to 0 for all wavelengths
                
                    # Fills the model dictionary
                    self.files.append(file+"_"+g)
                    self.models_dict[file+"_"+g] = {}
                    self.models_dict[file+"_"+g]['lin_wl'] = model[1].data['WAVELENGTH']
                    self.models_dict[file+"_"+g]['flux']   = model[1].data[g]

                    # Gets the parameters from the file name
                    Teff = int(file.split("_")[1].split('.')[0])
                    logg = int(g[-2]) + float(g[-1])/10
                    
                    FeH_sign = - 1 if (file[2] == 'm') else + 1
                    FeH  = FeH_sign * (int(file.split("_")[0][3]) + float(file.split("_")[0][4])/10)  

                    aFe = 0.4           

                    # Appends, to the template array, the stellar parameters of the model
                    self.templates.append([Teff, logg, FeH, aFe])


#====================================================================================================================#


class NGSL(SpectralModel):

    """
    This class is a child class of Model, it uses all the original methods from it, but it has a Load function to
    load the data from the model files.

    The models to be used with this class are described on:
        Gregg M. D., et al., 2006, in Koekemoer A. M., Goudfrooĳ P., Dressel L. L.,
            eds, The 2005 HST Calibration Workshop: Hubble After the Transition
            to Two-Gyro Mode. p. 209
        Heap S. R., Lindler D. J., 2007, in Vallenari A., Tantalo R., Portinari L.,
            Moretti A., eds, Astronomical Society of the Pacific Conference Series
            Vol. 374, From Stars to Galaxies: Building the Pieces to Build Up the
            Universe. p. 409

    Methods:
    ------------
    self.load(path: str):
        Loads the wavelenghts, fluxes and stellar parameters from the models folder passed as argument.
    """

    def load(self, path):

        # Looks into the directory passed for the templates
        file_list = os.listdir(path)

        # Initializes all the containers
        self.models_dict = { }
        self.templates   = [ ]
        self.files       = [ ]
        
        print()     
        for i, file in enumerate(file_list):

            # Reads the model
            model = fits.open(path+file)

            # Removes the extension from the name of the file
            file  = file.replace('.fits', '')

            # Message to the user
            message = f'Reading model \u001b[38;5;204m{i + 1}\u001b[0m of \u001b[38;5;204m{len(file_list)}\u001b[0m'
            sys.stdout.write(f'\u001b[32;1m>>>\u001b[0m {message} : \u001b[33;1m{file}\u001b[0m\r')
            sys.stdout.flush()

            # Fills the model dictionary
            self.files.append(file)
            self.models_dict[file] = {}
            self.models_dict[file]['lin_wl'] = model[1].data['WAVELENGTH']
            self.models_dict[file]['log_wl'] = np.log10(model[1].data['WAVELENGTH'])
            self.models_dict[file]['flux']   = model[1].data['FLUX']

            # Set the stellar parameters to 99 (undefined)
            Teff = 99
            logg = 99
        
            FeH  = 99
            aFe  = 99

            # Appends, to the template array, the stellar parameters of the model
            self.templates.append([Teff, logg, FeH, aFe])


#====================================00================================================================================#


# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                               sed_fitting.py
# ******************************************************************************

"""
Functions related to the SED fitting process used by the calibration pipeline
to predict S-PLUS magnitudes

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
improt sed_fitting as sf
----------------
"""


################################################################################
# Import external packages
import os.path
import sys
import numpy as np
import pandas as pd
from time import time
from astropy.table import Table

################################################################################
# Define functions

def get_chi2(model_ref_mag_array, ref_mag_array, ref_magerr_array):

    # Shift the models
    shift_array = ref_mag_array - model_ref_mag_array
    shift_array = shift_array.mean(axis = 1).reshape(-1,1)

    shifted_model_ref_mag_array = model_ref_mag_array + shift_array

    chi2 = (shifted_model_ref_mag_array - ref_mag_array) ** 2 / ref_magerr_array

    # Sum over all colors
    chi2 = chi2.sum(axis=1)

    return chi2


# *************************************************************
#   Absolute shift for model mags
# *************************************************************
def get_mag_shift(model_ref_mag_array, best_model_id, ref_mag_array):
    bestmod_ref_mag_array = model_ref_mag_array[best_model_id, :]
    delta_mags = ref_mag_array - bestmod_ref_mag_array

    mag_shift = np.mean(delta_mags)

    return mag_shift


# *************************************************************
#   Evaluate best model
# *************************************************************
def get_best_model(models, data, ref_mag_cols, pred_mag_cols=None,
                   bayesian = False):

    # Slice models array to get only reference magnitudes
    model_ref_mag_array = models.loc[:, ref_mag_cols].values

    # Slice data array to get only reference magnitudes
    ref_mag_array = data.loc[ref_mag_cols].values

    # Slice data array to get only reference magnitudes errors
    ref_magerr_cols = get_ref_magerr_cols(ref_mag_cols)
    ref_magerr_array = data.loc[ref_magerr_cols].values

    # Only used for the output
    if pred_mag_cols is not None:
        # Slice data array to get only S-PLUS magnitudes
        pred_mag_array = data.loc[pred_mag_cols].values

        # Slice data array to get only S-PLUS magnitudes errors
        pred_magerr_cols = get_ref_magerr_cols(pred_mag_cols)
        pred_magerr_array = data.loc[pred_magerr_cols].values

    # Calculate chi2 for each model
    chi2 = get_chi2(model_ref_mag_array=model_ref_mag_array,
                    ref_mag_array=ref_mag_array,
                    ref_magerr_array=ref_magerr_array)

    # Get the best model id
    if bayesian:
        prior = models['prior'].values
        posterior = [prior[i] * np.exp(-chi2[i]/2) for i in range(len(chi2))]

        best_model_id = np.argmax(posterior)

    else:
        best_model_id = np.argmin(chi2)

    # Calculate chi2 for each model
    mag_shift = get_mag_shift(model_ref_mag_array=model_ref_mag_array,
                              best_model_id=best_model_id,
                              ref_mag_array=ref_mag_array)

    # Get best model
    best_model = models.iloc[best_model_id, :]

    # Generate the output list
    output_list = [best_model.loc['model_id']]
    output_list += [best_model.loc['Teff']]
    output_list += [best_model.loc['logg']]
    output_list += [best_model.loc['FeH']]
    output_list += [best_model.loc['aFe']]
    output_list += [best_model.loc['EB_V']]
    output_list += [chi2[best_model_id]]
    output_list += [mag_shift]

    # Add reference magnitudes to the output
    for mag_ref, magerr_ref in zip(ref_mag_array, ref_magerr_array):
        output_list += [mag_ref]
        output_list += [magerr_ref]

    #if pred_mag_cols is not None:
    #    # Add splus magnitudes to the output
    #    for mag_ref, magerr_ref in zip(pred_mag_array, pred_magerr_array):
    #        output_list += [mag_ref]
    #        output_list += [magerr_ref]

    # Add to-predict magnitudes to the output
    if pred_mag_cols is not None:
        for i in range(len(pred_mag_cols)):
            pred_mag_col = pred_mag_cols[i]
            if pred_mag_col not in ref_mag_cols:
                output_list += [pred_mag_array[i]]
                output_list += [pred_magerr_array[i]]

    # Add model predicted magnitudes to the output, applying the shift
    for model_mag_ref in ref_mag_cols:
        output_list += [best_model.loc[model_mag_ref] + mag_shift]

    if pred_mag_cols is not None:
        for i in range(len(pred_mag_cols)):
            model_mag_pred = pred_mag_cols[i]
            if model_mag_pred not in ref_mag_cols:
                output_list += [best_model.loc[model_mag_pred] + mag_shift]

    # turn output into array
    output = np.array(output_list)

    return output


# *************************************************************
#   Load models
# *************************************************************
def load_models(models_file):
    # must load the file with the models and put it in a pd dataframe
    models = pd.read_csv(models_file, delim_whitespace=True, escapechar='#')
    models.columns = models.columns.str.replace(' ', '')

    return models

# *************************************************************
#   Load data
# *************************************************************


def load_data(data_file):

    # If necessary, transform from fits table to data frame
    if os.path.splitext(data_file)[1] == '.fits':
        fits_data = Table.read(data_file, format='fits')
        ref_data  = fits_data.to_pandas()

    else:
        ref_data = pd.read_csv(data_file, delim_whitespace=True, escapechar='#',
                               skipinitialspace=True)

    ref_data.columns = ref_data.columns.str.replace(' ', '')

    return ref_data


# *************************************************************
#   Get colnames for the errors
# *************************************************************
def get_ref_magerr_cols(ref_mag_cols):
    ref_magerr_cols = []

    for col in ref_mag_cols:
        ref_magerr_cols.append(col + '_err')

    return ref_magerr_cols


# *************************************************************
#   Fit model mags and save to a file
# *************************************************************
def get_model_mags(models_file, data_file, save_file,
                   ref_mag_cols, pred_mag_cols=None,
                   bayesian=False, ebv_cut=None):
    """
    Fit model mags to a reference catalog and predict the values of magnitudes
    for another filter system


    Parameters
    ----------
    models_file: str
        path to file with model magnitudes

    data_file: str
        path to file with reference magnitudes

    save_file: str
        path to the output file

    ref_mag_cols: tuple
        list of magnitudes names to be used to fit the models

    pred_mag_cols: tuple
        list of magnitudes names to be predicted by the models.

    bayesian: bool
        If true, best model is selected from the maximization of the posterior,
        otherwise, minimization of chi2.

    ebv_cut: float
        if not None, only models with EB_V = EBV_cut will be considered

    Returns
    -------
    Generates output file with predicted magnitudes

    """

    t0 = time()

    # load models
    print('Loading models from file %s' % models_file)
    models = load_models(models_file)

    if ebv_cut is not None:
        filt = models['EB_V'] == ebv_cut
        models = models[filt]

    # load data
    print('Loading data from file %s' % data_file)
    data = load_data(data_file)

    print('\nReference magnitudes being used are:')
    print(ref_mag_cols)

    ref_magerr_cols = get_ref_magerr_cols(ref_mag_cols)

    if pred_mag_cols is not None:
        pred_magerr_cols = get_ref_magerr_cols(pred_mag_cols)
        print('\nMagnitudes being predicted are:')
        print(pred_mag_cols)

    # Create output array

    Nlines = data.shape[0]
    Ncols = 4  # RA, DEC, X, Y
    Ncols += 8  # model_id, Teff, logg, FeH, aFe, EB_V, chi2, mag_shift
    Ncols += 3 * len(ref_mag_cols)  # mag_ref, mag_ref_err, mag_ref_model

    if pred_mag_cols is not None:
        for i in range(len(pred_mag_cols)):
            mag_name = pred_mag_cols[i]
            if mag_name not in ref_mag_cols:
                Ncols += 3

    output = np.full((Nlines, Ncols), np.nan)

    print('\n\nStarting to fit best model for each star ')

    # Obtain model mags for each star in data
    for i in range(Nlines):
        sys.stdout.write('\rFinding best model for star {0} of {1}'.format(i + 1,
                                                                           Nlines))
        sys.stdout.flush()

        # Put Model data in the output array
        output[i, 4:] = get_best_model(models=models,
                                       data=data.iloc[i, :],
                                       ref_mag_cols=ref_mag_cols,
                                       pred_mag_cols=pred_mag_cols,
                                       bayesian=bayesian)

    print('\n\nFinished estimating best model for {0} stars'.format(Nlines))

    output[:, 0] = data.loc[:, 'RAJ2000'].values
    output[:, 1] = data.loc[:, 'DEJ2000'].values
    output[:, 2] = data.loc[:, 'X'].values
    output[:, 3] = data.loc[:, 'Y'].values

    with open(save_file, 'w') as f:
        f.write('# RAJ2000 DEJ2000 X Y model_id Teff logg FeH aFe EB_V chi2 model_mag_shift')
        fmt = ['%.6f', '%.6f', '%.6f', '%.6f', '%d', '%d', '%.1f', '%.2f', "%.2f", '%.3f', '%.3f', '%.3f']

        for mag_name, mag_err_name in zip(ref_mag_cols, ref_magerr_cols):
            fmt += ['%.3f', '%.3f']
            f.write(' {0} {1}'.format(mag_name, mag_err_name))

        if pred_mag_cols is not None:
            #for mag_name, mag_err_name in zip(pred_mag_cols, pred_magerr_cols):
            #    fmt += ['%.3f', '%.3f']
            #    f.write(' {0} {1}'.format(mag_name, mag_err_name))
            for i in range(len(pred_mag_cols)):
                mag_name = pred_mag_cols[i]
                mag_err_name = pred_magerr_cols[i]
                if mag_name not in ref_mag_cols:
                    fmt += ['%.3f', '%.3f']
                    f.write(' {0} {1}'.format(mag_name, mag_err_name))

        for mag_name in ref_mag_cols:
            fmt += ['%.3f']
            f.write(' %s_mod' % mag_name)

        if pred_mag_cols is not None:
            for i in range(len(pred_mag_cols)):
                mag_name = pred_mag_cols[i]
                if mag_name not in ref_mag_cols:
                    fmt += ['%.3f']
                    f.write(' %s_mod' % mag_name)

        f.write('\n')
        np.savetxt(f, output, fmt=fmt)

    print('Results saved in file %s' % save_file)

    dt = time() - t0
    print ("\nThe minchi2 model magnitudes script took "
           "{0} seconds to find the best model for {1} stars".format(dt,
                                                                     Nlines))

    return output


# *************************************************************
#   save model fitting data
# *************************************************************
def save_data(output, save_file, ref_mag_cols, ref_magerr_cols,
              splus_mag_cols, splus_magerr_cols):

    with open(save_file, 'w') as f:
        f.write('# RA DEC model_id EB_V chi2')
        fmt = ['%.6f', '%.6f', '%d', '%.3f', '%.3f']

        for mag_name, mag_err_name in zip(ref_mag_cols, ref_magerr_cols):
            fmt += ['%.3f', '%.3f']
            f.write(' {0} {1}'.format(mag_name, mag_err_name))

        if splus_mag_cols is not None:
            for mag_name, mag_err_name in zip(splus_mag_cols, splus_magerr_cols):
                fmt += ['%.3f', '%.3f']
                f.write(' {0} {1}'.format(mag_name, mag_err_name))

        for mag_name in ref_mag_cols:
            fmt += ['%.3f']
            f.write(' %s_mod' % mag_name)

        if splus_mag_cols is not None:
            for mag_name in splus_mag_cols:
                fmt += ['%.3f']
                f.write(' %s_mod' % mag_name)

        f.write('\n')

        np.savetxt(f, output, fmt=fmt)



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                              fit_inst_offsets.py
#      Compares two instrumental photometries to derive zero-point offsets
# ******************************************************************************

"""
--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
This script should be run individually and not within the pipeline.py script

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 fit_zp_offsets.py *field_list_file* *calib1_path*

----------------
"""

################################################################################
# Import external packages

import os
import sys

addsteps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(addsteps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field_list_file = sys.argv[1]
calib_path = sys.argv[2]

save_path = os.path.join(calib_path, f'zp_offset_inst')
ut.makedir(save_path)

path_to_stilts = os.path.join(spluscalib_path, "external_programs",
                              "stilts.jar")

filters = ["SPLUS_U", "SPLUS_F378", "SPLUS_F395", "SPLUS_F410",
           "SPLUS_F430", "SPLUS_G", "SPLUS_F515", "SPLUS_R",
           "SPLUS_F660", "SPLUS_I", "SPLUS_F861", "SPLUS_Z"]

################################################################################
# Initiate log file

log_file_name = os.path.join(save_path, 'fit_inst_offsets.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(save_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")


################################################################################
# Reading field list

ut.printlog("*********** Reading field list **********", log_file)

fields = ut.load_field_list(field_list_file)

ut.printlog(f"Running the script for fields:", log_file)
ut.printlog(f"{fields}", log_file)

################################################################################
# Begin script


# ***************************************************
#    Crossmatching photometry files
# ***************************************************

def crossmatch_dual_psf_master_photometry():

    print("")
    ut.printlog(('********** '
                 'Crossmatching dual mode and psf master photometries '
                 '**********'),
                log_file)
    print("")

    for field in fields:

        save_name = f"{field}_master_photometry_dual_psf.fits"
        save_file = os.path.join(save_path, save_name)

        if not os.path.exists(save_file):
            dual_name = f"{field}_master_photometry_only_dual.fits"
            dual_file = os.path.join(calib_path, field, "Photometry", "dual",
                                     "master", dual_name)

            psf_name = f"{field}_master_photometry_only_psf.fits"
            psf_file = os.path.join(calib_path, field, "Photometry", "psf",
                                     "master", psf_name)

            values1 = "RAJ2000 DEJ2000"
            values2 = "RAJ2000 DEJ2000"

            cmd  = f"java -jar {path_to_stilts} tmatch2 "
            cmd += f"matcher=sky join=1and2 params=3 "
            cmd += f"in1={dual_file} ifmt1=fits values1='{values1}' "
            cmd += f"in2={psf_file} ifmt2=fits values2='{values2}' "
            cmd += f"out={save_file} ofmt=fits"

            ut.printlog(cmd, log_file)
            os.system(cmd)

        else:
            ut.printlog(f"File {save_file} already exists.", log_file)


crossmatch_dual_psf_master_photometry()


# ***************************************************
#    Estimating field dual_psf offsets
# ***************************************************

def estimating_dual_psf_offsets_by_field():

    print("")
    ut.printlog(('********** '
                 'Estimating dual/psf offsets for each field '
                 '**********'),
                log_file)
    print("")

    for field in fields:

        master_name = f"{field}_master_photometry_dual_psf.fits"
        master_file = os.path.join(save_path, master_name)

        save_name = f"{field}_offset_dual_psf.zp"
        save_file = os.path.join(save_path, save_name)

        if not os.path.exists(save_file):
            ut.estimate_field_dual_psf_offset(catalog   = master_file,
                                              save_file = save_file,
                                              filters   = filters)

            ut.printlog(f"Created file {save_file}", log_file)

        else:
            ut.printlog(f"File {save_file} already exists.", log_file)


estimating_dual_psf_offsets_by_field()


# ***************************************************
#    Estimating field dual_psf offsets
# ***************************************************

def estimating_overall_dual_psf_offsets():

    print("")
    ut.printlog(('********** '
                 'Overall dual/psf offsets'
                 '**********'),
                log_file)
    print("")

    save_name = f"all_offset_dual_psf.zp"
    save_file = os.path.join(save_path, save_name)

    if not os.path.exists(save_file):
        offset_files = []

        for field in fields:
            field_offset_name = f"{field}_offset_dual_psf.zp"
            field_offset_file = os.path.join(save_path,
                                             field_offset_name)

            offset_files.append(field_offset_file)

        ut.estimate_overall_dual_psf_offset(offset_files = offset_files,
                                            save_file = save_file,
                                            filters = filters)

    else:
        ut.printlog(f"File {save_file} already exists.", log_file)


estimating_overall_dual_psf_offsets()



# -*- coding: utf-8 -*-

# ******************************************************************************
#                          S-PLUS CALIBRATION PIPELINE
#                              correction_aper.py
#          Creates aperture corrected magnitudes in SExtractor catalogs
# ******************************************************************************


"""
Estimates and applies the aperture correction for the 3-arcsec diameter fixed
circular aperture.

Aperture corrections are obtained from the growth curve of each filter, derived
from the 32 fixed circular apertures given in the sextractor config file

The S-PLUS field is given as the first command line argument.

The set of filters, aperture to be corrected, cut in signal-to-noise and cut
in SExtractor CLASS_STAR parameter must be set in the configuration file
given as the second command line argument

--------------------------------------------------------------------------------
   FUNCTIONS:
--------------------------------------------------------------------------------

create_apercorr_paths()
obtain_single_mode_aper_corr()
plot_single_mode_growth_curve()
apply_single_mode_aperture_correction()
obtain_dual_mode_aper_corr()
plot_dual_mode_growth_curve()
apply_dual_mode_aperture_correction()

--------------------------------------------------------------------------------
   COMMENTS:
--------------------------------------------------------------------------------
Ideally this script should only be run through the pipeline.py script.

Assumes that photometry_single.py and/or photometry_dual.py (and optionally
correction_xy.py) has already been run for this field.

--------------------------------------------------------------------------------
   USAGE:
--------------------------------------------------------------------------------
$python3 correction_aper.py *field_name* *config_file*

----------------
"""

################################################################################
# Import external packages

import os
import sys

steps_path = os.path.split(__file__)[0]
pipeline_path = os.path.split(steps_path)[0]
spluscalib_path = os.path.split(pipeline_path)[0]

sys.path.append(spluscalib_path)

################################################################################
# Import spluscalib packages

from spluscalib import utils as ut

################################################################################
# Read parameters

field     = sys.argv[1]
conf_file = sys.argv[2]

conf = ut.pipeline_conf(conf_file)

################################################################################
# Get directories

field_path           = os.path.join(conf['run_path'], field)
photometry_path      = os.path.join(field_path, 'Photometry')

dual_path            = os.path.join(photometry_path, 'dual')
dual_catalogs_path   = os.path.join(dual_path, 'catalogs')
dual_xycorr_path     = os.path.join(dual_path, 'xy_correction')
dual_apercorr_path   = os.path.join(dual_path, 'aper_correction')
dual_sexconf_path    = os.path.join(dual_path, 'sexconf')

single_path          = os.path.join(photometry_path, 'single')
single_catalogs_path = os.path.join(single_path, 'catalogs')
single_xycorr_path   = os.path.join(single_path, 'xy_correction')
single_apercorr_path = os.path.join(single_path, 'aper_correction')
single_sexconf_path  = os.path.join(single_path, 'sexconf')

log_path             = os.path.join(photometry_path, 'logs')

################################################################################
# Initiate log file

ut.makedir(log_path)

log_file_name = os.path.join(log_path, 'aper_correction.log')
log_file_name = ut.gen_logfile_name(log_file_name)
log_file = os.path.join(photometry_path, log_file_name)

with open(log_file, "w") as log:
    log.write("")


################################################################################
# Log configuration

ut.printlog("PSF photometry parameters:", log_file)

single_mode_params = ['apercorr_diameter', 'apercorr_s2ncut',
                      'apercorr_starcut', 'apercorr_max_aperture']

for param in single_mode_params:
    try:
        ut.printlog(f"{param}: {conf[param]}", log_file)
    except KeyError:
        ut.printlog(f"{param}: NONE", log_file)

################################################################################
# Begin script

# ***************************************************
#    Create paths
# ***************************************************


def create_apercorr_paths():

    print("")
    ut.printlog(('********** '
                 'Generating aperture corrections paths '
                 '**********'),
                 log_file)
    print("")

    # Create single mode photometry path
    if 'photometry_dual' in conf['run_steps']:
        ut.makedir(dual_apercorr_path)
        ut.makedir(os.path.join(dual_apercorr_path, 'growth_curves'))

    if 'photometry_single' in conf['run_steps']:
        ut.makedir(single_apercorr_path)
        ut.makedir(os.path.join(single_apercorr_path, 'growth_curves'))


create_apercorr_paths()


# ***************************************************
#    Single Mode Aperture Correction
# ***************************************************

def obtain_single_mode_aper_corr():

    print("")
    ut.printlog(('********** '
                 'Estimating single mode aperture corrections '
                 '**********'),
                 log_file)
    print("")

    # Generate aperture correction file name
    aper_corr_name = f"{field}_apercorr_single.zp"
    save_file = os.path.join(single_apercorr_path, aper_corr_name)

    if not os.path.exists(save_file):

        # create file
        with open(save_file + 'temp', 'w') as f:
            f.write("")

        # Estimate aperture correction for each filter
        for filt in conf['filters']:

            # Get photometry catalog
            if conf['sex_XY_correction']:
                catalog_name = f"sex_{field}_{filt}_single_xycorr.fits"
                catalog = os.path.join(single_xycorr_path, catalog_name)
            else:
                catalog_name = f"sex_{field}_{filt}_single.fits"
                catalog = os.path.join(single_catalogs_path, catalog_name)

            # Get filt sexconfig file
            conf_file_name = f"{field}_{filt}_single.sex"
            sexconf_file = os.path.join(single_sexconf_path, conf_file_name)

            # Calculate filter aperture correction
            ut.obtain_aperture_correction(catalog=catalog,
                                          filt=filt,
                                          sexconf=sexconf_file,
                                          save_file=save_file + 'temp',
                                          aperture=conf['apercorr_diameter'],
                                          s2ncut=conf['apercorr_s2ncut'],
                                          starcut=conf['apercorr_starcut'],
                                          verbose=True,
                                    max_aperture=conf['apercorr_max_aperture'])

            ut.printlog((f"Estimated aperture correction for field {field}, "
                         f"filter {filt} (single mode)"), log_file)

    else:
        ut.printlog((f"Aperture correction already estimated for field "
                     f"{field} (single mode)"), log_file)

    # After finishing, remove temp file
    if not os.path.exists(save_file):
        cmd = f"mv {save_file+'temp'} {save_file}"
        ut.printlog(cmd, log_file)
        os.system(cmd)


if 'photometry_single' in conf['run_steps']:
    obtain_single_mode_aper_corr()


# ***************************************************
#    Plot growth curve - single mode
# ***************************************************


def plot_single_mode_growth_curve():

    print("")
    ut.printlog(('********** '
                 'Plotting single mode growth curves '
                 '**********'),
                 log_file)
    print("")

    for filt in conf['filters']:

        plot_name = f'growth_curve_{field}_{filt}_single.png'
        save_file = os.path.join(single_apercorr_path, 'growth_curves',
                                 plot_name)

        if not os.path.exists(save_file):

            # Get photometry catalog
            if conf['sex_XY_correction']:
                catalog_name = f"sex_{field}_{filt}_single_xycorr.fits"
                catalog = os.path.join(single_xycorr_path, catalog_name)
            else:
                catalog_name = f"sex_{field}_{filt}_single.fits"
                catalog = os.path.join(single_catalogs_path, catalog_name)

            # Get filt sexconfig file
            conf_file_name = f"{field}_{filt}_single.sex"
            sexconf_file = os.path.join(single_sexconf_path, conf_file_name)

            # Calculate filter aperture correction
            ut.growth_curve_plotter(catalog=catalog,
                                    filt=filt,
                                    sexconf=sexconf_file,
                                    save_file=save_file,
                                    aperture=conf['apercorr_diameter'],
                                    s2ncut=conf['apercorr_s2ncut'],
                                    starcut=conf['apercorr_starcut'],
                                    verbose=True,
                                    max_aperture=conf['apercorr_max_aperture'])

            ut.printlog((f"Plotted growth curve for field {field}, "
                         f"filter {filt} (single mode)"), log_file)

        else:
            ut.printlog((f"Growth curve already plotted for field "
                         f"{field}, filter {filt}  (single mode)"), log_file)


if 'photometry_single' in conf['run_steps']:
    plot_single_mode_growth_curve()


# ***************************************************
#    Apply aperture correction - single mode
# ***************************************************


def apply_single_mode_aperture_correction():

    print("")
    ut.printlog(('********** '
                 'Applying single mode aperture correction '
                 '**********'),
                 log_file)
    print("")

    aper_corr_name = f"{field}_apercorr_single.zp"
    aper_corr_file = os.path.join(single_apercorr_path, aper_corr_name)

    for filt in conf['filters']:

        save_name = f'sex_{field}_{filt}_single_apercorr.fits'
        save_file = os.path.join(single_apercorr_path, save_name)

        if not os.path.exists(save_file):

            # Get photometry catalog
            if conf['sex_XY_correction']:
                catalog_name = f"sex_{field}_{filt}_single_xycorr.fits"
                catalog = os.path.join(single_xycorr_path, catalog_name)
            else:
                catalog_name = f"sex_{field}_{filt}_single.fits"
                catalog = os.path.join(single_catalogs_path, catalog_name)

            # Get filt sexconfig file
            conf_file_name = f"{field}_{filt}_single.sex"
            sexconf_file = os.path.join(single_sexconf_path, conf_file_name)

            # Calculate filter aperture correction
            ut.apply_aperture_correction(catalog       = catalog,
                                         filt          = filt,
                                         sexconf       = sexconf_file,
                                         aperture = conf['apercorr_diameter'],
                                         apercorr_file = aper_corr_file,
                                         save_file     = save_file,
                                         field         = field,
                                         drname   = conf['data_release_name'],
                                         sexmode       = 'single')

            ut.printlog((f"Applied aperture correction for field {field}, "
                         f"filter {filt} (single mode)"), log_file)

        else:
            ut.printlog((f"Aperture correction already applied for field "
                         f"{field}, filter {filt} (single mode)"), log_file)


if 'photometry_single' in conf['run_steps']:
    apply_single_mode_aperture_correction()


# ***************************************************
#    Dual Mode Aperture Correction
# ***************************************************

def obtain_dual_mode_aper_corr():

    print("")
    ut.printlog(('********** '
                 'Estimating dual mode aperture corrections '
                 '**********'),
                 log_file)
    print("")

    # Generate aperture correction file name
    aper_corr_name = f"{field}_apercorr_dual.zp"
    save_file = os.path.join(dual_apercorr_path, aper_corr_name)

    if not os.path.exists(save_file):

        # create file
        with open(save_file + 'temp', 'w') as f:
            f.write("")

        # Estimate aperture correction for each filter
        for filt in conf['filters']:

            # Get photometry catalog
            if conf['sex_XY_correction']:
                catalog_name = f"sex_{field}_{filt}_dual_xycorr.fits"
                catalog      = os.path.join(dual_xycorr_path, catalog_name)
            else:
                catalog_name = f"sex_{field}_{filt}_dual.fits"
                catalog      = os.path.join(dual_catalogs_path, catalog_name)

            # Get filt sexconfig file
            conf_file_name = f"{field}_{filt}_dual.sex"
            sexconf_file   = os.path.join(dual_sexconf_path, conf_file_name)

            # Calculate filter aperture correction
            ut.obtain_aperture_correction(catalog   = catalog,
                                          filt      = filt,
                                          sexconf   = sexconf_file,
                                          save_file = save_file+'temp',
                                          aperture  = conf['apercorr_diameter'],
                                          s2ncut    = conf['apercorr_s2ncut'],
                                          starcut   = conf['apercorr_starcut'],
                                          verbose=True,
                                     max_aperture=conf['apercorr_max_aperture'])

            ut.printlog((f"Estimated aperture correction for field {field}, "
                         f"filter {filt} (dual mode)"), log_file)

    else:
        ut.printlog((f"Aperture correction already estimated for field "
                     f"{field} (dual mode)"), log_file)

    # After finishing, remove temp file
    if not os.path.exists(save_file):
        cmd = f"mv {save_file+'temp'} {save_file}"
        ut.printlog(cmd, log_file)
        os.system(cmd)


if 'photometry_dual' in conf['run_steps']:
    obtain_dual_mode_aper_corr()


# ***************************************************
#    Plot growth curve - dual mode
# ***************************************************


def plot_dual_mode_growth_curve():

    print("")
    ut.printlog(('********** '
                 'Plotting dual mode growth curves '
                 '**********'),
                 log_file)
    print("")

    for filt in conf['filters']:

        plot_name = f'growth_curve_{field}_{filt}_dual.png'
        save_file = os.path.join(dual_apercorr_path, 'growth_curves',
                                 plot_name)

        if not os.path.exists(save_file):

            # Get photometry catalog
            if conf['sex_XY_correction']:
                catalog_name = f"sex_{field}_{filt}_dual_xycorr.fits"
                catalog = os.path.join(dual_xycorr_path, catalog_name)
            else:
                catalog_name = f"sex_{field}_{filt}_dual.fits"
                catalog = os.path.join(dual_catalogs_path, catalog_name)

            # Get filt sexconfig file
            conf_file_name = f"{field}_{filt}_dual.sex"
            sexconf_file = os.path.join(dual_sexconf_path, conf_file_name)

            # Calculate filter aperture correction
            ut.growth_curve_plotter(catalog=catalog,
                                    filt=filt,
                                    sexconf=sexconf_file,
                                    save_file=save_file,
                                    aperture=conf['apercorr_diameter'],
                                    s2ncut=conf['apercorr_s2ncut'],
                                    starcut=conf['apercorr_starcut'],
                                    verbose=True,
                                    max_aperture=conf['apercorr_max_aperture'])

            ut.printlog((f"Plotted growth curve for field {field}, "
                         f"filter {filt} (dual mode)"), log_file)

        else:
            ut.printlog((f"Growth curve already plotted for field "
                         f"{field}, filter {filt}  (dual mode)"), log_file)


if 'photometry_dual' in conf['run_steps']:
    plot_dual_mode_growth_curve()


# ***************************************************
#    Apply aperture correction - dual mode
# ***************************************************


def apply_dual_mode_aperture_correction():

    print("")
    ut.printlog(('********** '
                 'Applying dual mode aperture correction '
                 '**********'),
                 log_file)
    print("")

    aper_corr_name = f"{field}_apercorr_dual.zp"
    aper_corr_file = os.path.join(dual_apercorr_path, aper_corr_name)

    for filt in conf['filters']:

        save_name = f'sex_{field}_{filt}_dual_apercorr.fits'
        save_file = os.path.join(dual_apercorr_path, save_name)

        if not os.path.exists(save_file):

            # Get photometry catalog
            if conf['sex_XY_correction']:
                catalog_name = f"sex_{field}_{filt}_dual_xycorr.fits"
                catalog = os.path.join(dual_xycorr_path, catalog_name)
            else:
                catalog_name = f"sex_{field}_{filt}_dual.fits"
                catalog = os.path.join(dual_catalogs_path, catalog_name)

            # Get filt sexconfig file
            conf_file_name = f"{field}_{filt}_dual.sex"
            sexconf_file = os.path.join(dual_sexconf_path, conf_file_name)

            # Calculate filter aperture correction
            ut.apply_aperture_correction(catalog       = catalog,
                                         filt          = filt,
                                         sexconf       = sexconf_file,
                                         aperture = conf['apercorr_diameter'],
                                         apercorr_file = aper_corr_file,
                                         save_file     = save_file,
                                         field         = field,
                                         drname = conf['data_release_name'],
                                         sexmode       = 'dual')

            ut.printlog((f"Applied aperture correction for field {field}, "
                         f"filter {filt} (dual mode)"), log_file)

        else:
            ut.printlog((f"Aperture correction already applied for field "
                         f"{field}, filter {filt} (dual mode)"), log_file)


if 'photometry_dual' in conf['run_steps']:
    apply_dual_mode_aperture_correction()
